%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Definition of external functions of module IO:
%
% Note: the Prolog term '$stream'('$inoutstream'(In,Out)) represents a handle
% for a stream that is both readable (on In) and writable (on Out)
% Otherwise, handles are represented by Prolog terms of the form '$stream'(N).
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- use_module('../prologbasics').
:- use_module(library(system)).
:- ensure_loaded(user:prim_ports). % to implement IO.prim_hWaitForInputsOrMsg

?- block prim_stdin(?,-,?).
prim_stdin(Stream,E0,E) :- prolog_flag(user_input,Stream), E0=E.

?- block prim_stdout(?,-,?).
prim_stdout(Stream,E0,E) :- prolog_flag(user_output,Stream), E0=E.

?- block prim_stderr(?,-,?).
prim_stderr(Stream,E0,E) :- prolog_flag(user_error,Stream), E0=E.

?- block prim_openFile(?,?,?,?,-,?).
prim_openFile(RA,RMode,_,'$io'(Stream),E0,E) :-
	derefAll(RA,A), deref(RMode,Mode),
	string2Atom(A,FName),
	curryFileMode2plmode(Mode,PMode),
	open(FName,PMode,Stream), E0=E.

curryFileMode2plmode('IO.ReadMode',read).
curryFileMode2plmode('IO.WriteMode',write).
curryFileMode2plmode('IO.AppendMode',append).

?- block prim_hClose(?,?,?,-,?).
prim_hClose(Stream,W,Result,E0,E) :-
	hnf(Stream,HStream,E0,E1),
	prim_hClose_exec(HStream,W,Result,E1,E).

?- block prim_hClose_exec(-,?,?,?,?), prim_hClose_exec(?,?,?,-,?).
prim_hClose_exec('$stream'('$inoutstream'(In,Out)),_,'$io'('()'),E0,E) :- !,
	flush_output(Out),
	close(Out),
	(In==Out -> true ; close(In)),
	E0=E.
prim_hClose_exec(Stream,_,'$io'('()'),E0,E) :-
	(isOutputStream(Stream) -> flush_output(Stream) ; true),
	close(Stream), E0=E.


?- block prim_hFlush(?,?,?,-,?).
prim_hFlush(Stream,W,Result,E0,E) :-
        hnf(Stream,HStream,E0,E1),
	prim_hFlush_exec(HStream,W,Result,E1,E).

?- block prim_hFlush_exec(-,?,?,?,?), prim_hFlush_exec(?,?,?,-,?).
prim_hFlush_exec('$stream'('$inoutstream'(_,Out)),_,'$io'('()'),E0,E) :- !,
	flush_output(Out), E0=E.
prim_hFlush_exec(Stream,_,'$io'('()'),E0,E) :-
	(isOutputStream(Stream) -> flush_output(Stream) ; true),
	E0=E.


?- block prim_hIsEOF(?,?,?,-,?).
prim_hIsEOF(Stream,W,Result,E0,E) :-
        hnf(Stream,HStream,E0,E1),
	prim_hIsEOF_exec(HStream,W,Result,E1,E).

?- block prim_hIsEOF_exec(-,?,?,?,?), prim_hIsEOF_exec(?,?,?,-,?).
prim_hIsEOF_exec('$stream'('$inoutstream'(In,_)),_,'$io'(B),E0,E) :- !,
	(at_end_of_stream(In) -> B='True' ; B='False'),
	E0=E.
prim_hIsEOF_exec(Stream,_,'$io'(B),E0,E) :-
	(at_end_of_stream(Stream) -> B='True' ; B='False'),
	E0=E.


?- block prim_hSeek(?,?,?,?,?,-,?).
prim_hSeek(RHandle,RSeekMode,RPos,_,'$io'('()'),E0,E) :-
	deref(RHandle,Handle),
	deref(RSeekMode,SeekMode),
	deref(RPos,Pos),
	currySeekMode2plmode(SeekMode,PlSM),
	seek(Handle,Pos,PlSM,_),
	E0=E.

currySeekMode2plmode('IO.AbsoluteSeek',bof).
currySeekMode2plmode('IO.RelativeSeek',current).
currySeekMode2plmode('IO.SeekFromEnd',eof).


?- block prim_hWaitForInputs(?,?,?,?,-,?).
prim_hWaitForInputs(RStreams,RTO,_,'$io'(N),E0,E) :-
	derefAll(RStreams,Streams),
	selectInstreams(Streams,InStreams),
	deref(RTO,TimeOut),
	(TimeOut<0 -> TO=off
	            ; TOSec is TimeOut//1000,
	              TOMSec is (TimeOut mod 1000) * 1000, % in milliseconds
	              TO = TOSec:TOMSec),
	waitForInputDataOnStreams(InStreams,TO,N),
	!, E0=E.

selectInstreams([],[]).
selectInstreams(['$stream'('$inoutstream'(In,_))|Streams],[In|InStreams]) :- !,
	selectInstreams(Streams,InStreams).
selectInstreams([Stream|Streams],[Stream|InStreams]) :-
	selectInstreams(Streams,InStreams).


?- block prim_hGetChar(?,?,?,-,?).
prim_hGetChar(Stream,W,Result,E0,E) :-
        hnf(Stream,HStream,E0,E1),
	prim_hGetChar_exec(HStream,W,Result,E1,E).

?- block prim_hGetChar_exec(-,?,?,?,?), prim_hGetChar_exec(?,?,?,-,?).
prim_hGetChar_exec('$stream'('$inoutstream'(In,_)),_,'$io'(C),E0,E) :- !,
	get_code(In,N), char_int(C,N),
	E0=E.
prim_hGetChar_exec(Stream,_,'$io'(C),E0,E) :-
	get_code(Stream,N), char_int(C,N),
	E0=E.

?- block prim_hPutChar(?,?,?,?,-,?).
prim_hPutChar(Stream,RC,_,R,E0,E) :-
	hnf(Stream,HStream,E0,E1),
	hnf(RC,C,E1,E2),
	prim_hPutChar_exec(HStream,C,R,E2,E).

?- block prim_hPutChar_exec(-,?,?,?,?), prim_hPutChar_exec(?,-,?,?,?),
         prim_hPutChar_exec(?,?,?,-,?).
prim_hPutChar_exec('$stream'('$inoutstream'(_,Out)),C,'$io'('()'),E0,E) :- !,
	char_int(C,N), put_code(Out,N), E0=E.
prim_hPutChar_exec(Stream,C,'$io'('()'),E0,E) :-
	char_int(C,N), put_code(Stream,N), E0=E.


?- block prim_hIsReadable(?,?,?,-,?).
prim_hIsReadable(Stream,W,Result,E0,E) :-
        hnf(Stream,HStream,E0,E1),
	prim_hIsReadable_exec(HStream,W,Result,E1,E).

?- block prim_hIsReadable_exec(-,?,?,?,?), prim_hIsReadable_exec(?,?,?,-,?).
prim_hIsReadable_exec('$stream'('$inoutstream'(_,_)),_,'$io'('True'),E0,E) :- !,
	E0=E.
prim_hIsReadable_exec(Stream,_,'$io'(B),E0,E) :-
	(isInputStream(Stream) -> B='True' ; B='False'),
	E0=E.


?- block prim_hIsWritable(?,?,?,-,?).
prim_hIsWritable(Stream,W,Result,E0,E) :-
        hnf(Stream,HStream,E0,E1),
	prim_hIsWritable_exec(HStream,W,Result,E1,E).

?- block prim_hIsWritable_exec(-,?,?,?,?), prim_hIsWritable_exec(?,?,?,-,?).
prim_hIsWritable_exec('$stream'('$inoutstream'(_,_)),_,'$io'('True'),E0,E) :- !,
	E0=E.
prim_hIsWritable_exec(Stream,_,'$io'(B),E0,E) :-
	(isOutputStream(Stream) -> B='True' ; B='False'),
	E0=E.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% choice on a stream and an external port message stream:
?- block prim_hWaitForInputsOrMsg(?,-,?,?,?,?),
         prim_hWaitForInputsOrMsg(?,?,?,?,-,?).
prim_hWaitForInputsOrMsg(Handles,share(M),World,Result,E0,E) :-
        !,
	get_mutable(V,M),
	(V='$eval'(R) % external message stream has been already evaluated
	 -> E0=E, Result='$io'('Right'(R))
	  ; prim_hWaitForInputsOrMsg(Handles,V,World,CResult,E0,E),
	    (CResult='$io'('Left'(_))
  	     -> Result=CResult
	      ; CResult='$io'('Right'(S)),
	        (compileWithSharing(variable) -> propagateShare(S,TResult) ; S=TResult),
	        Result='$io'('Right'(TResult)),
	        update_mutable('$eval'(TResult),M))).
prim_hWaitForInputsOrMsg(_,[M|Ms],_,'$io'('Right'([M|Ms])),E0,E) :-
	!, E0=E. % stream already evaluated
prim_hWaitForInputsOrMsg(RHandles,[],_,'$io'('Left'(N)),E0,E) :- !,
	% message stream is empty, so anything must be received from the handles.
	derefAll(RHandles,Handles),
	selectInstreams(Handles,InStreams),
	waitForInputDataOnStreams(InStreams,off,N),
	!, E0=E.
prim_hWaitForInputsOrMsg(Handles,'Ports.basicServerLoop'(Port),World,
			Result,E0,E) :-
        Port='Ports.internalPort'(_,_,PNr,_),
	checkIncomingPortStreams(PNr,InPortStream,OutPortStream),
	!,
	readStreamUntilEndOfTerm(InPortStream,MsgString),
	(parse_received_message(InPortStream,OutPortStream,MsgString,Msg)
	 -> ifTracePort((write(user_error,'TRACEPORTS: Received message: '),
		         write(user_error,Msg), nl(user_error))),
	    Result='$io'('Right'([Msg|'Ports.basicServerLoop'(Port)])), E0=E
	  ; write(user_error,'ERROR: Illegal message received (ignored): '),
	    putChars(user_error,MsgString), nl(user_error),
	    prim_hWaitForInputsOrMsg(Handles,'Ports.basicServerLoop'(Port),World,
	                                   Result,E0,E)),
	!.
prim_hWaitForInputsOrMsg(RHandles,'Ports.basicServerLoop'(Port),World,Result,E0,E):-
	derefAll(RHandles,Handles),
	selectInstreams(Handles,InStreams),
        Port='Ports.internalPort'(_,_,PNr,Socket),
	waitForSocketOrInputStreams(Socket,Client,InPortStream,OutPortStream,
				    InStreams,Index),
	(Client=no
	 -> % there is input on Handles:
	    Result='$io'('Left'(Index)), E0=E
	  ; % there is a client connection:
	    ifTracePort((write(user_error,'TRACEPORTS: Connection to client: '),
		         write(user_error,Client), nl(user_error))),
	    (readPortMessage(PNr,InPortStream,OutPortStream,MsgString)
	     -> (parse_received_message(InPortStream,OutPortStream,MsgString,Msg)
	         -> ifTracePort((write(user_error,'TRACEPORTS: Received message: '),
		                 write(user_error,Msg), nl(user_error))),
		    Result='$io'('Right'([Msg|'Ports.basicServerLoop'(Port)])), E0=E
	          ; write(user_error,'ERROR: Illegal message received (ignored): '),
		    putChars(user_error,MsgString), nl(user_error),
		    prim_hWaitForInputsOrMsg(Handles,
                                 'Ports.basicServerLoop'(Port),World,Result,E0,E))
	      ; % try reading next message:
		prim_hWaitForInputsOrMsg(Handles,
		                 'Ports.basicServerLoop'(Port),World,Result,E0,E))),
	!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
